@@ -1,4 +1,4 @@
-use rustpython_bytecode::bytecode::{CodeObject, Instruction, Label, Location};
+use rustpython_bytecode::bytecode::{CodeFlags, CodeObject, Instruction, Label, Location};
 
 pub trait OutputStream: From<CodeObject> + Into<CodeObject> {
     /// Output an instruction
@@ -7,6 +7,8 @@ pub trait OutputStream: From<CodeObject> + Into<CodeObject> {
     fn set_label(&mut self, label: Label);
     /// Mark the inner CodeObject as a generator
     fn mark_generator(&mut self);
+    /// Check to see if the inner CodeObject is a generator
+    fn is_generator(&self) -> bool;
 }
 
 pub struct CodeObjectStream {
@@ -34,6 +36,9 @@ impl OutputStream for CodeObjectStream {
         self.code.label_map.insert(label, position);
     }
     fn mark_generator(&mut self) {
-        self.code.is_generator = true;
+        self.code.flags |= CodeFlags::IS_GENERATOR;
+    }
+    fn is_generator(&self) -> bool {
+        self.code.flags.contains(CodeFlags::IS_GENERATOR)
     }
 }
